;****************************************************************
;	     HuMusic- Music / BGM Script Decoding		*
;****************************************************************
; Music is stored in two data structures, songs and track
; packages (collections of songs). While playback of individual
; songs is possible, it is recommended that all songs be combined
; into a track package for simplicity.
;
; The data structures used for music have no bank boundary
; limitations and may begin, end, and wrap anywhere in an 8KB/PCE or
; 256KB/MCGenjin bank.
;
;
; ---- TRACK PACKAGES ----
; The current track package's base address is stored in HuMusic_PackBase,
; and must reside in ROM. A well behaved track package will be formatted
; as shown below:
; 
; Offset	Contents
; -------	---------
; 0		"T"
; 1		"R"
; 2		"K"
; 3		Number of Songs / Tracks
; 4+
;  \
;   0-3		Song Start Addresses (Relative Lo, Hi, Bank, Map)
; (4 * (NumTracks + 1))+
;  \
;   Song Data
;
;
; -------- SONGS ---------
; The current song's base address is stored in HuMusic_SongBase, which
; will automatically change when using the track package playback functions.
; Note that while Track Packages (and their included songs) are required
; to fall in the MCGenjin swappable ROM region, individual songs may be
; placed arbitrarily (even in EXRAM). Like a track package, a song also
; begins with a header whose format is given below:
;
; Offset	Contents
; -------	---------
; 0		"M"
; 1		"U"
; 2		"S"
; 3		Number of Instrument Definitions
; 4-7		Base Address of Instrument Table (Relative Lo, Hi, Bank, Map)
; 8+		Channel Script Descriptors:
;  \
;   0			"C"
;   1-4			Start Address (Relative Lo, Hi, Bank, Map)
;   5			Initial Priority (0 if the channel is unused)
; (8 + (6 * 6))+
;  \
;   Music Data
;
;
; Notable Functions
; -----------------
; -HuMusic_TrackReq:
;	Request playback of a music track whose number is in A.
; -HuMusic_Play:
;	Begin playing the song whose base address is in HuMusic_SongBase.
; -HuMusic_Stop:
;	Ceases the irritating noise of any currently playing music.
;
; -HuMusic_ReqAtten:
;	Set the music attenuation from 0 (min, loudest) to 15 (max, silent).
;
;
; Music Script Commands
;-----------------------
.define HUMUSIC_SET_PRIORITY	0
.define HUMUSIC_SET_PANNING	1
.define HUMUSIC_DELAY		2

.define HUMUSIC_LOOP_START	3
.define HUMUSIC_LOOP_END	4
.define HUMUSIC_CALL		5
.define HUMUSIC_RETURN		6
.define HUMUSIC_BREAK		7

.define HUMUSIC_PITCH		8
.define HUMUSIC_NOTE_ON		9
.define HUMUSIC_NOTE_OFF	10
.define HUMUSIC_SAMPLE		11


;****************************************************************
; HuMusic_TrackReq:						*
;	Request playback of the music track in A (in our track	*
; package), which will begin on the next HuSound driver tick.	*
;****************************************************************
.section "HuMusic_TrackReq" force
HuMusic_TrackReq:
	STA HuMusic_ReqTrack
	LDA #$FF
	STA HuMusic_ReqPlay
	rts
.ends
;****************************************************************
; HuMusic_TrackPlay:						*
;	Begin playback of the music track in A (from our track	*
; package), out of range tracks will be ignored. This should	*
; only be called inside the HuMusic driver update tick, user	*
; requests for track playback must use HuMusic_TrackReq above.	*
;****************************************************************
.section "HuMusic_TrackPlay" force
HuMusic_TrackPlay:
	TAY
	LDA HuMusic_NumTracks	; Check to make sure that a valid track pack
	bne _packOpen		; already been opened and the requested track
_packOpen			; is within range of its contents.
	CPY HuMusic_NumTracks
	bcc _setupTrack
	rts

_setupTrack			; Begin setup of the track, first calculating
	TYA			; its directory entry offset (4-byte addr each).
	STA HuMusic_SongBase + LO
	STZ HuMusic_SongBase + HI
	STZ HuMusic_SongBase + BANK
	STZ HuMusic_SongBase + MCMAP
	clc
	ROL HuMusic_SongBase + LO
	ROL HuMusic_SongBase + HI
	ROL HuMusic_SongBase + LO
	ROL HuMusic_SongBase + HI

				; Calculate the entry's true address relative to
				; the directory base (result in HuSound_CurStream).
	sei
	MCG_Rel HuMusic_DirBase HuMusic_SongBase (HuSound_CurStream + ADDR_RAM)

				; Since HuSound_CurStream now points to the base of
				; our directory entry, we can just use HuSound_ReadByte
	jsr HuSound_ReadByte	; to copy out the entry data (our song's base address).
	STA HuMusic_SongBase + LO
	jsr HuSound_ReadByte
	STA HuMusic_SongBase + HI
	jsr HuSound_ReadByte
	STA HuMusic_SongBase + BANK
	jsr HuSound_ReadByte
	STA HuMusic_SongBase + MCMAP

				; Then resolve this address relative to the pack base,
				; storing the result in HuMusic_SongBase.

	MCG_Rel HuMusic_PackBase HuMusic_SongBase HuMusic_SongBase

				; We're done evaluating the song's true base address
				; and can drop down into HuMusic_Play which will also
.ends				; stop any currently playing song.




;****************************************************************
; HuMusic_Play:							*
;	Cease playback on the current music track (if playing),	*
; then play the song whose base address is in HuMusic_SongBase.	*
; Note that this takes quite a few cycles and disables IRQs.	*
;****************************************************************
.section "HuMusic_Play" force
HuMusic_Play:
	jsr HuMusic_Stop

	LDA MCG_LastMap		; Before we start, save our current mapping
	PHA			; configuration and then copy our song's base
	TMA #MPR4		; address to the stream decoder's address.
	PHA			; This way we can easily traverse the header,
	PHX			; since we're disabling IRQs as well we don't
	PHY			; have to worry about conflicts.

	sei
	LDA HuMusic_SongBase + LO
	STA HuSound_CurStream + LO
	LDA HuMusic_SongBase + HI
	STA HuSound_CurStream + HI

	LDA HuMusic_SongBase + BANK
	STA HuSound_CurStream + BANK
	LDA HuMusic_SongBase + MCMAP
	STA HuSound_CurStream + MCMAP

	CLY			; Ensure the music track has a valid three
_checkHeader			; byte identifier ("MUS"). If not, abort the
	jsr HuSound_ReadByte	; setup operation immediately.
	CMP _validHeader,Y
	beq _musLetterOkay
	jmp _done
_musLetterOkay
	INY
	CPY #3
	bne _checkHeader


	jsr HuSound_ReadByte	; ---- INSTRUMENT TABLE SETUP ----
	STA HuMusic_NumInst	; Copy the number of instruments in the
	CLY			; table and then resolve its true base address
_copyInstOffset			; (relative to the start of the song).
	jsr HuSound_ReadByte
	STA HuMusic_InstBase,Y
	INY
	CPY #4
	bne _copyInstOffset

	MCG_Rel HuMusic_SongBase HuMusic_InstBase HuMusic_InstBase

	CLX
_setupChannels			; ---- MUSIC SCRIPTS SETUP ----
	jsr HuSound_ReadByte	; Each channel has its own 6-Byte header, the
	CMP #'C'		; first of which should be the character 'C'.
	bne _done

	CLY			; Following the lead character is the relative
_copyScriptOffset		; script address, which we'll need to copy and
	jsr HuSound_ReadByte	; resolve as with the instrument table.
	STA HuSound_RelAddr,Y
	INY
	CPY #4
	bne _copyScriptOffset

	MCG_Rel HuMusic_SongBase HuSound_RelAddr HuSound_RelAddr

	LDY.w HuSound_DecOffs,X
	PHX
	CLX
_copyScriptResolved
	LDA HuSound_RelAddr,X
	STA HuMusic_CurStream,Y
	INY
	INX
	CPX #4
	bne _copyScriptResolved
	PLX

	jsr HuSound_ReadByte	; Last is the channel's starting priority, which
	STA HuMusic_Priority,X	; can just be copied directly.

	INX
	CPX #HUSOUND_CHANNELS
	bne _setupChannels

	LDA #$FF
	STA HuMusic_isPlaying	; Enable music playback in the driver update routine.
_done
	PLA			; Pop'n, but not Twinbee
	TAM #MPR4
	PLA
	STA MCGENJIN_BANK
	PLY
	PLX

	cli
	rts

_validHeader	.db "MUS"
.ends
;****************************************************************
; HuMusic_Stop:							*
;	Stop any currently playing music and free its channels,	*
; the music decoding state variables will also be reset.	*
;****************************************************************
.section "HuMusic_Stop" force
HuMusic_Stop:
	LDA HuMusic_isPlaying
	bne _killMusic

	LDX #(HUSOUND_CHANNELS - 1)
_resetParams
	STZ HuMusic_Priority,X	; <- Even if no music is playing, make sure
	STZ HuMusic_Delay,X	; to reset the major state variables.
	STZ HuMusic_StackLevel,X

	STZ HuSound_FreqPitch_Lo,X
	STZ HuSound_FreqPitch_Hi,X
	STZ HuSound_FreqPitch_De,X
	STZ HuSound_FreqPitchAdj_Lo,X
	STZ HuSound_FreqPitchAdj_Hi,X
	STZ HuSound_FreqPitchAdj_De,X
	DEX
	bpl _resetParams
	rts


_killMusic
	PHX
	sei
	LDX #(HUSOUND_CHANNELS - 1)
_killSamples
	LDA HuMusic_Priority,X	; Before stopping the music itself, kill off
	CMP HuSample_Priority,X	; any samples it may have instantiated.
	beq _isActive
	DEX
	bmi _doneSamples
_isActive
	CLA
	sei
	STA HuSample_Playing,X
	STA HuSound_noWriteback,X
	STA HuSample_Priority,X
	cli
	DEX
	bpl _killSamples
_doneSamples

	LDX #(HUSOUND_CHANNELS - 1)
_freeChannels
	LDA HuSound_isInstrument,X
	beq _notMuChannel	; <- Each of the channels must be checked
				; to see if it is occupied by an instrument,
	STZ HuSound_Priority,X	; and disabled immediately if so.
	STZ HuSound_isInstrument,X
_notMuChannel
	STZ HuMusic_Priority,X	; <- Reset the PRIORITY, DELAY, and STACK
	STZ HuMusic_Delay,X	; for the next played music track.
	STZ HuMusic_StackLevel,X

	STZ HuSound_FreqPitch_Lo,X
	STZ HuSound_FreqPitch_Hi,X
	STZ HuSound_FreqPitch_De,X
	STZ HuSound_FreqPitchAdj_Lo,X
	STZ HuSound_FreqPitchAdj_Hi,X
	STZ HuSound_FreqPitchAdj_De,X
	DEX
	bpl _freeChannels

	STZ HuMusic_isPlaying
	cli
	PLX
	rts
.ends




;****************************************************************
; HuMusic_ReqAtten:						*
;	Request a change in the current music attenuation level	*
; where the new attenuation value is supplied in A and ranges	*
; from 0-15, where 0 is no attenuation (loudest) and 15 is	*
; maximum attenuation (silence).				*
;****************************************************************
.section "HuMusic_ReqAtten" force
HuMusic_ReqAtten:
	STA HuMusic_AttenLevel

	PHX
	LDX #(HUSOUND_CHANNELS - 1)
	LDA #$FF
_reqAdjusts
	STA HuMusic_AttenReq,X
	DEX
	bpl _reqAdjusts
	PLX
	rts
.ends
;****************************************************************
; HuMusic_ApplyAtten:						*
;	Subtract the current music attenuation level from the	*
; channel's panning and apply the results. This should only be	*
; called from within the HuMusic driver update tick. To change	*
; the attenuation level use HuMusic_ReqAtten, above.		*
;****************************************************************
.section "HuMusic_ApplyAtten" force
HuMusic_ApplyAtten:
	LDA HuMusic_Panning,X
	TAY
_adjVol_L
	LSR
	LSR
	LSR
	LSR
	sec
	SBC HuMusic_AttenLevel
	bpl _setVol_L

	CLA
_setVol_L
	ROL
	ROL
	ROL
	ROL
	AND #$F0
	STA HuMusic_VolFinal,X

_adjVol_R
	TYA
	AND #$0F
	sec
	SBC HuMusic_AttenLevel
	bpl _setVol_R

	CLA
_setVol_R
	ORA HuMusic_VolFinal,X
	STA HuMusic_VolFinal,X

	LDA HuSound_isInstrument,X
	bne _isInstrument	; See if the music track is currently
	LDA HuSound_Priority,X	; in control of this channel, if so-
	bne _notInstrument	; adjust the panning immediately.
				; Idle channels will also have their
_isInstrument			; panning adjusted for diagnostic use.
	LDA HuMusic_VolFinal,X
	STA HuSound_Panning,X
_notInstrument
	sei
	LDA HuSample_Playing,X	; If a sample is currently playing, see
	beq _notSample		; it it's one of ours...

	LDA HuMusic_Priority,X
	CMP HuSample_Priority,X
	bne _notSample

	STX PSG_CHANNEL
	LDA HuMusic_VolFinal,X
	STA PSG_PAN
_notSample
	cli
	rts
.ends




;****************************************************************
; HuMusic_Init:							*
;	Initialize the music playback software, and set up the	*
; track package (directory) at HuMusic_PackBase.		*
;****************************************************************
.section "HuMusic_Init" force
HuMusic_Init:
	STZ HuMusic_NumTracks	; Preset track count to zero.
	LDA HuMusic_PackBase + LO
	ORA HuMusic_PackBase + HI
	ORA HuMusic_PackBase + BANK
	ORA HuMusic_PackBase + MCMAP
	bne _checkPack		; Only investigate the track pack if
	rts			; is base address is nonzero.

_checkPack
	LDA MCG_LastMap		; Preserve our current mapping setup
	PHA			; and then copy our track package
	TMA #MPR4		; base address to the stream decoder
	PHA			; for simplified traversal.

	LDA HuMusic_PackBase + LO
	STA HuSound_CurStream + LO
	LDA HuMusic_PackBase + HI
	STA HuSound_CurStream + HI

	LDA HuMusic_PackBase + MCMAP
	STA HuSound_CurStream + MCMAP

	LDA HuMusic_PackBase + BANK
	CMP #BANK_ROM		; Track packs are only allowed to be
	bcc _shoveBank		; in ROM, and must be placed in the
	bra _checkHeader	; MCGenjin bankable area. So automatically
_shoveBank			; push the address up if we're in the
	ADC #BANK_ROM		; fixed ROM region below.

_checkHeader
	STA HuSound_CurStream + BANK

	CLY			; Look for the track pack "TRK" header,
_checkLetter			; aborting the setup process if any byte
	jsr HuSound_ReadByte	; mismatch is found.
	CMP _validHeader,Y
	bne _done
	INY
	CPY #3
	bne _checkLetter

	jsr HuSound_ReadByte	; Header is okay, so just read in the track
	STA HuMusic_NumTracks	; count and set our current read offset as the
				; directory base.

	LDA HuSound_CurStream + LO
	STA HuMusic_DirBase + LO
	LDA HuSound_CurStream + HI
	STA HuMusic_DirBase + HI
	LDA HuSound_CurStream + BANK
	STA HuMusic_DirBase + BANK
	LDA HuSound_CurStream + MCMAP
	STA HuMusic_DirBase + MCMAP

_done
	PLA
	TAM #MPR4
	PLA
	STA MCGENJIN_BANK
	rts

_validHeader	.db "TRK"
.ends




;****************************************************************
; HuMusic_Decode:						*
;	Step through the channel's music script to dispatch	*
; note ons, offs, and other useful things. Should be called	*
; every driver tick only if music playback is enabled and the	*
; channel's music priority is greater than zero.		*
;****************************************************************
.section "HuMusic_Decode" force
HuMusic_Decode:
	LDY.w HuSound_DecOffs,X
	LDA HuMusic_CurStream,Y
	STA HuSound_CurStream + LO
	INY
	LDA HuMusic_CurStream,Y
	STA HuSound_CurStream + HI
	INY
	LDA HuMusic_CurStream,Y
	STA HuSound_CurStream + BANK
	INY
	LDA HuMusic_CurStream,Y
	STA HuSound_CurStream + MCMAP

	LDA HuMusic_AttenReq,X	; See if we got any attenuation update
	beq _checkDelay		; requests before doing any script work...

	PHY
	jsr HuMusic_ApplyAtten
	PLY
	STZ HuMusic_AttenReq,X
_checkDelay
	LDA HuMusic_Delay,X	; If decoding is stalled on this channel
	beq _getNextCommand	; (rest or wait between commands), just
				; decrement the wait counter and leave.
	DEC HuMusic_Delay,X
	bra _decodeDone		; <- Exit point for decode process
_getNextCommand
	jsr HuSound_ReadByte	; Fetch our next command byte,
	TAY			; then dispatch its respective handler.
	LDA _musCommandLo,Y
	STA HuSound_JumpDest + LO
	LDA _musCommandHi,Y
	STA HuSound_JumpDest + HI
	jsr HuSound_VarJump + ADDR_RAM

	bra _checkDelay		; Go back and check the delay value to
_decodeDone			; see if its been set (so we can stop).

	LDY.w HuSound_DecOffs,X
	LDA HuSound_CurStream + LO
	STA HuMusic_CurStream,Y
	INY
	LDA HuSound_CurStream + HI
	STA HuMusic_CurStream,Y
	INY
	LDA HuSound_CurStream + BANK
	STA HuMusic_CurStream,Y
	INY
	LDA HuSound_CurStream + MCMAP
	STA HuMusic_CurStream,Y
	rts

;---- Low Command Table ----------------------------------------
_musCommandLo	.db <HuMusic_Com_SetPri,<HuMusic_Com_SetPan,<HuMusic_Com_Del

		.db <HuMusic_Com_SetLoop,<HuMusic_Com_Loop
		.db <HuMusic_Com_Call,<HuMusic_Com_Ret,<HuMusic_Com_Break

		.db <HuMusic_Com_Pitch
		.db <HuMusic_Com_NoteOn,<HuMusic_Com_NoteOff
		.db <HuMusic_Com_Sample
;---- High Command Table ---------------------------------------
_musCommandHi	.db >HuMusic_Com_SetPri,>HuMusic_Com_SetPan,>HuMusic_Com_Del

		.db >HuMusic_Com_SetLoop,>HuMusic_Com_Loop
		.db >HuMusic_Com_Call,>HuMusic_Com_Ret,>HuMusic_Com_Break

		.db >HuMusic_Com_Pitch
		.db >HuMusic_Com_NoteOn,>HuMusic_Com_NoteOff
		.db >HuMusic_Com_Sample
.ends
;****************************************************************
; Music Script Commands						*
;****************************************************************
; HUMUSIC_SET_PRIORITY(Priority)
; Set the music track's current playback priority. Setting the
; priority to zero will terminate the decode process.
;----------------------------------------------------------------
.section "HuMusic_Com_SetPri" force
HuMusic_Com_SetPri:
	jsr HuSound_ReadByte	; If our priority was just set to zero,
	STA HuMusic_Priority,X	; the music on this channel is finished. So
	bne _noKill		; we can kill any instrument if it is active
				; and then bail out of the decode process.
	LDA HuSound_isInstrument,X
	bne _noActiveInst

	STZ HuSound_Priority,X
_noActiveInst
	PLA
	PLA
_noKill
	rts
.ends

; HUMUSIC_SET_PANNING(Panning)
; Set the music track's current stereo panning. The one byte
; panning value is in the format %LLLLRRRR, where larger values
; indicate a larger amplitude. So a pan of $F0 would yield
; the loudest volume through the left speaker.
;----------------------------------------------------------------
.section "HuMusic_Com_SetPan" force
HuMusic_Com_SetPan:
	jsr HuSound_ReadByte	; Woah, hold it! Don't set that panning
	STA HuMusic_Panning,X	; value immediately. We've got to adjust it
	jmp HuMusic_ApplyAtten	; based upon the current music atten level.
.ends

; HUMUSIC_DELAY(Delay)
; Delay the track's decoding for the specified count in ticks.
;----------------------------------------------------------------
.section "HuMusic_Com_Del" force
HuMusic_Com_Del:
	jsr HuSound_ReadByte
	STA HuMusic_Delay,X
	rts
.ends

; HUMUSIC_LOOP_START(Loop Count)
; Set the start of a looping region and define how many counts
; the loop will last. Negative loop counts are considered infinite.
;----------------------------------------------------------------
.section "HuMusic_Com_SetLoop" force
HuMusic_Com_SetLoop:
	jsr HuSound_ReadByte

	clc			; Copy the provided loop count...
	LDA HuMusic_StackLevel,X
	ADC.w HuSound_LoopOffs,X
	TAY
	LDA HuSound_CurByte
	STA HuMusic_LoopCount,Y

	clc			; ...then push our current decoding address
	LDY HuMusic_StackLevel,X
	LDA.w HuSound_StackOffs,X
	ADC HuSound_DecOffs,Y
	TAY
	LDA HuSound_CurStream + LO
	STA HuMusic_DecoStack,Y
	INY
	LDA HuSound_CurStream + HI
	STA HuMusic_DecoStack,Y
	INY
	LDA HuSound_CurStream + BANK
	STA HuMusic_DecoStack,Y
	INY
	LDA HuSound_CurStream + MCMAP
	STA HuMusic_DecoStack,Y	; So when we LOOP later, all we have to do
				; is pull this value (next command byte).
	INC HuMusic_StackLevel,X
	rts
.ends

; HUMUSIC_LOOP_END()
; Decrement the current loop count, branching back to the loop start
; if the loop count is greater than zero or negative (infinite).
;----------------------------------------------------------------
.section "HuMusic_Com_Loop" force
HuMusic_Com_Loop:
	LDA HuMusic_StackLevel,X
	bne _stackNotEmpty	; <- Only attempt to LOOP if our stack has
	rts			; at least ONE entry in it, otherwise abort.

_stackNotEmpty
	clc
	DEA
	ADC.w HuSound_LoopOffs,X
	TAY
	LDA HuMusic_LoopCount,Y
	bmi _restoreAddr	; <- Negative loop counts are infinite / always.
	beq _closeLoop		; <- Counts of zero never loop / act as single-shot.

	LDA HuMusic_LoopCount,Y
	DEA
	STA HuMusic_LoopCount,Y
	bne _restoreAddr
_closeLoop
	DEC HuMusic_StackLevel,X
	rts

_restoreAddr
	LDY HuMusic_StackLevel,X
	DEY
	clc
	LDA.w HuSound_StackOffs,X
	ADC HuSound_DecOffs,Y
	TAY
	LDA HuMusic_DecoStack,Y
	STA HuSound_CurStream + LO
	INY
	LDA HuMusic_DecoStack,Y
	STA HuSound_CurStream + HI
	INY
	LDA HuMusic_DecoStack,Y
	STA HuSound_CurStream + BANK
	INY
	LDA HuMusic_DecoStack,Y
	STA HuSound_CurStream + MCMAP
	rts
.ends

; HUMUSIC_CALL(Dest Lo, Dest Hi, Dest Bank, Dest Map)
; Call ("jsr") the music script at the supplied relative address.
; The target address is equal to Dest + HuMusic_SongBase.
;----------------------------------------------------------------
.section "HuMusic_Com_Call" force
HuMusic_Com_Call:
	jsr HuSound_ReadByte
	STA HuSound_TarStream + LO
	jsr HuSound_ReadByte
	STA HuSound_TarStream + HI
	jsr HuSound_ReadByte
	STA HuSound_TarStream + BANK
	jsr HuSound_ReadByte
	STA HuSound_TarStream + MCMAP

	clc
	LDA HuMusic_StackLevel,X
	ADC.w HuSound_LoopOffs,X
	TAY			; CALLs are pushed on the stack as a special
	LDA #$80		; case loop with a count of $80 (-128).
	STA HuMusic_LoopCount,Y

	clc
	LDY HuMusic_StackLevel,X
	LDA.w HuSound_StackOffs,X
	ADC HuSound_DecOffs,Y
	TAY
	LDA HuSound_CurStream + LO
	STA HuMusic_DecoStack,Y
	INY
	LDA HuSound_CurStream + HI
	STA HuMusic_DecoStack,Y
	INY
	LDA HuSound_CurStream + BANK
	STA HuMusic_DecoStack,Y
	INY
	LDA HuSound_CurStream + MCMAP
	STA HuMusic_DecoStack,Y

	INC HuMusic_StackLevel,X

	clc
	LDA HuSound_TarStream + LO
	ADC HuMusic_SongBase + LO
	STA HuSound_CurStream + LO

	LDA HuSound_TarStream + HI
	ADC HuMusic_SongBase + HI
	CMP #>ADDR_OBJ_HI
	bcc _writeHiAddress
	SBC #>(ADDR_OBJ_HI - ADDR_OBJ_LO)
	sec
_writeHiAddress
	STA HuSound_CurStream + HI

	LDA HuSound_TarStream + BANK
	ADC HuMusic_SongBase + BANK
	CMP #(BANK_ROM + MCBANK_PAGES)
	bcc _writeBank
	SBC #MCBANK_PAGES
	sec
_writeBank
	STA HuSound_CurStream + BANK

	LDA HuSound_TarStream + MCMAP
	ADC HuMusic_SongBase + MCMAP
	STA HuSound_CurStream + MCMAP
	rts
.ends

; HUMUSIC_RETURN()
; Return ("rts") from the current music script position to the
; command after the last CALL performed.
;----------------------------------------------------------------
.section "HuMusic_Com_Ret" force
HuMusic_Com_Ret:
	LDA HuMusic_StackLevel,X
	bne _stackNotEmpty
	rts	

_stackNotEmpty
	DEA
	STA HuMusic_StackLevel,X
	TAY
	clc
	LDA.w HuSound_StackOffs,X
	ADC HuSound_DecOffs,Y
	TAY

	LDA HuMusic_DecoStack,Y
	STA HuSound_CurStream + LO
	INY
	LDA HuMusic_DecoStack,Y
	STA HuSound_CurStream + HI
	INY
	LDA HuMusic_DecoStack,Y
	STA HuSound_CurStream + BANK
	INY
	LDA HuMusic_DecoStack,Y
	STA HuSound_CurStream + MCMAP
	rts
.ends

; HUMUSIC_BREAK()
; Return ("rts") all channels (not just the current one) to the
; lowest return address found in their pattern CALL stack,
; effectively bringing all channels back to their "main" script.
;----------------------------------------------------------------
.section "HuMusic_Com_Break" force
HuMusic_Com_Break:
	LDY.w HuSound_DecOffs,X
	LDA HuSound_CurStream + LO
	STA HuMusic_CurStream,Y
	INY
	LDA HuSound_CurStream + HI
	STA HuMusic_CurStream,Y
	INY
	LDA HuSound_CurStream + BANK
	STA HuMusic_CurStream,Y
	INY
	LDA HuSound_CurStream + MCMAP
	STA HuMusic_CurStream,Y	; Back up our current channel's decode pointer.

	PHX
	LDX #(HUSOUND_CHANNELS - 1)
_breakChannels
	LDA HuMusic_Priority,X	; Scan each active channel for its lowest 
	beq _skipChannel	; CALL, returning it to that decode address
	jsr _breakChannel	; and popping off all upper stack entries if found.
_skipChannel
	DEX
	bpl _breakChannels
	PLX

	LDY.w HuSound_DecOffs,X	; We're done, resume decoding at our original address
	LDA HuMusic_CurStream,Y	; or wherever we've RETURNed to.
	STA HuSound_CurStream + LO
	INY
	LDA HuMusic_CurStream,Y
	STA HuSound_CurStream + HI
	INY
	LDA HuMusic_CurStream,Y
	STA HuSound_CurStream + BANK
	INY
	LDA HuMusic_CurStream,Y
	STA HuSound_CurStream + MCMAP
	rts

_breakChannel			; <- Search n' RET
	LDA HuMusic_StackLevel,X
	bne _stackNotEmpty
	rts	

_stackNotEmpty			; Starting from the stack base, search
	CLA			; for the first entry with a loop count
_checkLevel			; of $80. Then pull its address if found.
	CMP HuMusic_StackLevel,X
	beq _noCalls
	PHA

	clc
	ADC.w HuSound_LoopOffs,X
	TAY
	LDA HuMusic_LoopCount,Y
	CMP #$80
	beq _foundCall
	PLA
	INA
	bra _checkLevel
_foundCall
	STZ HuMusic_Delay,X	; Breaking should kill any delay value,
	PLA			; channels with no calls are left alone.
	STA HuMusic_StackLevel,X
	TAY
				; This is a two-step pull, first copy the return
	clc			; address into the temporary stream decode pointer.
	LDA.w HuSound_StackOffs,X
	ADC HuSound_DecOffs,Y
	TAY
	LDA HuMusic_DecoStack,Y
	STA HuSound_CurStream + LO
	INY
	LDA HuMusic_DecoStack,Y
	STA HuSound_CurStream + HI
	INY
	LDA HuMusic_DecoStack,Y
	STA HuSound_CurStream + BANK
	INY
	LDA HuMusic_DecoStack,Y
	STA HuSound_CurStream + MCMAP

	LDY.w HuSound_DecOffs,X	; Then copy that back into the channel's decode pointer.
	LDA HuSound_CurStream + LO
	STA HuMusic_CurStream,Y
	INY
	LDA HuSound_CurStream + HI
	STA HuMusic_CurStream,Y
	INY
	LDA HuSound_CurStream + BANK
	STA HuMusic_CurStream,Y
	INY
	LDA HuSound_CurStream + MCMAP
	STA HuMusic_CurStream,Y
_noCalls
	rts
.ends


; HUMUSIC_PITCH(Pitch Lo, Hi, Pitch Adjust Lo, Hi, De)
; Set the current pitch offset and adjustment rate.
;----------------------------------------------------------------
.section "HuMusic_Com_Pitch" force
HuMusic_Com_Pitch:
	jsr HuSound_ReadByte
	STA HuSound_FreqPitch_Lo,X
	jsr HuSound_ReadByte
	STA HuSound_FreqPitch_Hi,X
	STZ HuSound_FreqPitch_De,X

	jsr HuSound_ReadByte
	STA HuSound_FreqPitchAdj_Lo,X
	jsr HuSound_ReadByte
	STA HuSound_FreqPitchAdj_Hi,X
	jsr HuSound_ReadByte
	STA HuSound_FreqPitchAdj_De,X
	rts
.ends

; HUMUSIC_NOTE_ON(Instrument, Frequency Lo, Hi)
; Start playing the specified intstrument at the given base frequency.
;----------------------------------------------------------------
.section "HuMusic_Com_NoteOn" force
HuMusic_Com_NoteOn:
	LDA HuMusic_Priority,X	; We can only dispatch a note if we're
	CMP HuSound_Priority,X	; of greater or equal priority to whatever
	bcs _checkInstrument	; is currently occupying the channel.

	jsr HuSound_ReadByte	; If we're not going to start playing the
	jsr HuSound_ReadByte	; instrument, just skip its three arguments
	jmp HuSound_ReadByte	; and head back to the decoding loop.

_checkInstrument
	jsr HuSound_ReadByte
	CMP HuMusic_NumInst
	bcc _setupInstrument

	jsr HuSound_ReadByte
	jmp HuSound_ReadByte

_setupInstrument		; Instrument Offset = 4 * Instrument Number
	STA HuMusic_InstOffset + LO
	STZ HuMusic_InstOffset + HI
	STZ HuMusic_InstOffset + BANK
	STZ HuMusic_InstOffset + MCMAP
	clc
	ROL HuMusic_InstOffset + LO
	ROL HuMusic_InstOffset + HI
	ROL HuMusic_InstOffset + LO
	ROL HuMusic_InstOffset + HI
				; Since we're going to be temporarily changing
				; our decode address (CurStream) when starting
				; the instrument, we'll read in the base frequency
				; now and push the resulting decode address
	jsr HuSound_ReadByte	; (now at the next command) to the stack.
	STA HuSound_FreqBase_Lo,X
	jsr HuSound_ReadByte
	STA HuSound_FreqBase_Hi,X
	STZ HuSound_FreqBase_De,X

	LDA HuSound_CurStream + LO
	PHA
	LDA HuSound_CurStream + HI
	PHA
	LDA HuSound_CurStream + BANK
	PHA
	LDA HuSound_CurStream + MCMAP
	PHA
				; ---- INSTRUMENT RESOLVE AND DISPATCH ----
	MCG_Rel HuMusic_InstBase HuMusic_InstOffset (HuSound_CurStream + ADDR_RAM)

	jsr HuSound_ReadByte
	STA HuMusic_InstOffset + LO
	jsr HuSound_ReadByte
	STA HuMusic_InstOffset + HI
	jsr HuSound_ReadByte
	STA HuMusic_InstOffset + BANK
	jsr HuSound_ReadByte
	STA HuMusic_InstOffset + MCMAP

	MCG_Rel HuMusic_InstBase HuMusic_InstOffset (HuSound_CurStream + ADDR_RAM)
	jsr HuSFX_StartInst

	PLA
	STA HuSound_CurStream + MCMAP
	PLA
	STA HuSound_CurStream + BANK
	PLA
	STA HuSound_CurStream + HI
	PLA
	STA HuSound_CurStream + LO
	rts
.ends

; HUMUSIC_NOTE_OFF()
; Have the currently playing instrument (if any) move to the
; note off portion of its script.
;----------------------------------------------------------------
.section "HuMusic_Com_NoteOff" force
HuMusic_Com_NoteOff:
	LDA HuSound_isInstrument,X
	bne _moveDecoder	; Make sure the music script is still in
	rts			; control of this channel...

_moveDecoder
	LDY.w HuSound_DecOffs,X	; ...if it is, then we can alter its
	LDA HuSFX_NoteOff,Y	; current decode address to that of
	STA HuSFX_CurStream,Y	; the note off portion of the script.
	INY
	LDA HuSFX_NoteOff,Y
	STA HuSFX_CurStream,Y
	INY
	LDA HuSFX_NoteOff,Y
	STA HuSFX_CurStream,Y
	INY
	LDA HuSFX_NoteOff,Y
	STA HuSFX_CurStream,Y

	STZ HuSFX_Delay,X	; Clear out any delays and reset the
	STZ HuSFX_StackLevel,X	; stack level. Please go NOW!
	rts
.ends

; HUMUSIC_SAMPLE(Sample Number)
; Start playing the specified PCM sample using HuSample.
;----------------------------------------------------------------
.section "HuMusic_Com_Sample" force
HuMusic_Com_Sample:
	jsr HuSound_ReadByte
	TAY

	sei			; Check to see if a sample is already playing...
	LDA HuSample_Playing,X
	beq _setupSample

	LDA HuMusic_Priority,X	; We only want to kick out a currently
	CMP HuSample_Priority,X	; playing sample if our priority is
	bcs _setupSample	; greater than or equal to.
	cli
	rts

_setupSample
	LDA HuMusic_Priority,X
	STA HuSample_Priority,X
				; Force the channel into DDA mode
	STX PSG_CHANNEL		; and set the current volume and panning.
	LDA #(PSG_CONTROL_SAMPLE | $1F)
	STA PSG_CONTROL		; We need to preemptively disable IRQs
	LDA HuMusic_VolFinal,X	; (even though they may be re-disabled
	STA PSG_PAN		; in HuSample_Play) to prevent the scenario
				; of the writeback routine triggering right
				; as we switch the channel to DDA, and putting
				; it back into wave or lfsr mode.
	TYA
	PHX
	SXY			; A = Sample #, Y = Channel
	jsr HuSample_Play	; Dispatch using HuSample
	cli
	PLX			; X = Channel (again)
	rts
.ends
